@@ -20,13 +20,12 @@
 
 from __future__ import absolute_import, division, print_function
 
-import ast
 import base64
 import contextlib
 import datetime
 import inspect
-import os
 import random as rnd_module
+import sys
 import traceback
 import warnings
 import zlib
@@ -91,7 +90,7 @@ from hypothesis.internal.reflection import (
 )
 from hypothesis.reporting import current_verbosity, report, verbose_report
 from hypothesis.searchstrategy.collections import TupleStrategy
-from hypothesis.searchstrategy.strategies import SearchStrategy
+from hypothesis.searchstrategy.strategies import MappedSearchStrategy, SearchStrategy
 from hypothesis.statistics import note_engine_for_statistics
 from hypothesis.utils.conventions import infer
 from hypothesis.version import __version__
@@ -214,18 +213,29 @@ def decode_failure(blob):
         )
 
 
-class WithRunner(SearchStrategy):
+class WithRunner(MappedSearchStrategy):
     def __init__(self, base, runner):
         assert runner is not None
-        self.base = base
+        MappedSearchStrategy.__init__(self, base)
         self.runner = runner
 
     def do_draw(self, data):
         data.hypothesis_runner = self.runner
-        return self.base.do_draw(data)
+        return self.mapped_strategy.do_draw(data)
+
+    def __repr__(self):
+        return "WithRunner(%r, runner=%r)" % (self.mapped_strategy, self.runner)
 
 
 def is_invalid_test(name, original_argspec, generator_arguments, generator_kwargs):
+    """Check the arguments to ``@given`` for basic usage constraints.
+
+    Most errors are not raised immediately; instead we return a dummy test
+    function that will raise the appropriate error if it is actually called.
+    When the user runs a subset of tests (e.g via ``pytest -k``), errors will
+    only be reported for tests that actually ran.
+    """
+
     def invalid(message):
         def wrapped_test(*arguments, **kwargs):
             raise InvalidArgument(message)
@@ -283,7 +293,7 @@ def is_invalid_test(name, original_argspec, generator_arguments, generator_kwarg
         repr(kw) for kw in original_argspec.kwonlyargs if kw not in generator_kwargs
     ]
     if missing:
-        raise InvalidArgument(
+        return invalid(
             "Missing required kwarg{}: {}".format(
                 "s" if len(missing) > 1 else "", ", ".join(missing)
             )
@@ -348,14 +358,7 @@ def get_random_for_wrapped_test(test, wrapped_test):
 
 
 def process_arguments_to_given(
-    wrapped_test,
-    arguments,
-    kwargs,
-    generator_arguments,
-    generator_kwargs,
-    argspec,
-    test,
-    settings,
+    wrapped_test, arguments, kwargs, generator_kwargs, argspec, test, settings,
 ):
     selfy = None
     arguments, kwargs = convert_positional_arguments(wrapped_test, arguments, kwargs)
@@ -397,59 +400,30 @@ def process_arguments_to_given(
     return arguments, kwargs, test_runner, search_strategy
 
 
-def run_once(fn):
-    """Wraps a no-args function so that its outcome is cached.
-    We use this for calculating various lists of exceptions
-    the first time we use them."""
-    result = [None]
-
-    def run():
-        if result[0] is None:
-            result[0] = fn()
-            assert result[0] is not None
-        return result[0]
-
-    run.__name__ = fn.__name__
-    return run
-
-
-@run_once
 def skip_exceptions_to_reraise():
     """Return a tuple of exceptions meaning 'skip this test', to re-raise.
 
     This is intended to cover most common test runners; if you would
-    like another to be added please open an issue or pull request.
+    like another to be added please open an issue or pull request adding
+    it to this function and to tests/cover/test_lazy_import.py
     """
-    import unittest
-
     # This is a set because nose may simply re-export unittest.SkipTest
-    exceptions = {unittest.SkipTest}
-
-    try:  # pragma: no cover
-        from unittest2 import SkipTest
-
-        exceptions.add(SkipTest)
-    except ImportError:
-        pass
-
-    try:  # pragma: no cover
-        from pytest.runner import Skipped
-
-        exceptions.add(Skipped)
-    except ImportError:
-        pass
-
-    try:  # pragma: no cover
-        from nose import SkipTest as NoseSkipTest
-
-        exceptions.add(NoseSkipTest)
-    except ImportError:
-        pass
-
+    exceptions = set()
+    # We use this sys.modules trick to avoid importing libraries -
+    # you can't be an instance of a type from an unimported module!
+    # This is fast enough that we don't need to cache the result,
+    # and more importantly it avoids possible side-effects :-)
+    if "unittest" in sys.modules:
+        exceptions.add(sys.modules["unittest"].SkipTest)
+    if "unittest2" in sys.modules:  # pragma: no cover
+        exceptions.add(sys.modules["unittest2"].SkipTest)
+    if "nose" in sys.modules:  # pragma: no cover
+        exceptions.add(sys.modules["nose"].SkipTest)
+    if "_pytest" in sys.modules:  # pragma: no branch
+        exceptions.add(sys.modules["_pytest"].outcomes.Skipped)
     return tuple(sorted(exceptions, key=str))
 
 
-@run_once
 def failure_exceptions_to_catch():
     """Return a tuple of exceptions meaning 'this test has failed', to catch.
 
@@ -457,12 +431,8 @@ def failure_exceptions_to_catch():
     like another to be added please open an issue or pull request.
     """
     exceptions = [Exception]
-    try:  # pragma: no cover
-        from _pytest.outcomes import Failed
-
-        exceptions.append(Failed)
-    except ImportError:
-        pass
+    if "_pytest" in sys.modules:  # pragma: no branch
+        exceptions.append(sys.modules["_pytest"].outcomes.Failed)
     return tuple(exceptions)
 
 
@@ -483,11 +453,6 @@ def new_given_argspec(original_argspec, generator_kwargs):
     )
 
 
-ROOT = os.path.dirname(__file__)
-
-STDLIB = os.path.dirname(os.__file__)
-
-
 class StateForActualGivenExecution(object):
     def __init__(self, test_runner, search_strategy, test, settings, random, had_seed):
         self.test_runner = test_runner
@@ -506,9 +471,19 @@ class StateForActualGivenExecution(object):
         self.files_to_propagate = set()
         self.failed_normally = False
 
-        self.used_examples_from_database = False
+    def execute_once(
+        self, data, print_example=False, is_final=False, expected_failure=None
+    ):
+        """Run the test function once, using ``data`` as input.
 
-    def execute(self, data, print_example=False, is_final=False, expected_failure=None):
+        If the test raises an exception, it will propagate through to the
+        caller of this method. Depending on its type, this could represent
+        an ordinary test failure, or a fatal error, or a control exception.
+
+        If this method returns normally, the test might have passed, or
+        it might have placed ``data`` in an unsuccessful state and then
+        swallowed the corresponding control exception.
+        """
         text_repr = [None]
         if self.settings.deadline is None:
             test = self.test
@@ -534,11 +509,12 @@ class StateForActualGivenExecution(object):
                 return result
 
         def run(data):
-            if not hasattr(data, "can_reproduce_example_from_repr"):
-                data.can_reproduce_example_from_repr = True
+            # Set up dynamic context needed by a single test run.
             with local_settings(self.settings):
                 with deterministic_PRNG():
                     with BuildContext(data, is_final=is_final):
+
+                        # Generate all arguments to the test function.
                         args, kwargs = data.draw(self.search_strategy)
                         if expected_failure is not None:
                             text_repr[0] = arg_string(test, args, kwargs)
@@ -548,10 +524,6 @@ class StateForActualGivenExecution(object):
                                 test.__name__,
                                 arg_string(test, args, kwargs),
                             )
-                            try:
-                                ast.parse(example)
-                            except SyntaxError:
-                                data.can_reproduce_example_from_repr = False
                             report("Falsifying example: %s" % (example,))
                         elif current_verbosity() >= Verbosity.verbose:
                             report(
@@ -560,7 +532,12 @@ class StateForActualGivenExecution(object):
                             )
                         return test(*args, **kwargs)
 
+        # Run the test function once, via the executor hook.
+        # In most cases this will delegate straight to `run(data)`.
         result = self.test_runner(data, run)
+
+        # If a failure was expected, it should have been raised already, so
+        # instead raise an appropriate diagnostic error.
         if expected_failure is not None:
             exception, traceback = expected_failure
             if (
@@ -593,9 +570,16 @@ class StateForActualGivenExecution(object):
             )
         return result
 
-    def evaluate_test_data(self, data):
+    def _execute_once_for_engine(self, data):
+        """Wrapper around ``execute_once`` that intercepts test failure
+        exceptions and single-test control exceptions, and turns them into
+        appropriate method calls to `data` instead.
+
+        This allows the engine to assume that any exception other than
+        ``StopTest`` must be a fatal error, and should stop the entire engine.
+        """
         try:
-            result = self.execute(data)
+            result = self.execute_once(data)
             if result is not None:
                 fail_health_check(
                     self.settings,
@@ -607,21 +591,35 @@ class StateForActualGivenExecution(object):
                     HealthCheck.return_value,
                 )
         except UnsatisfiedAssumption:
+            # An "assume" check failed, so instead we inform the engine that
+            # this test run was invalid.
             data.mark_invalid()
+        except StopTest:
+            # The engine knows how to handle this control exception, so it's
+            # OK to re-raise it.
+            raise
         except (
             HypothesisDeprecationWarning,
             FailedHealthCheck,
-            StopTest,
         ) + skip_exceptions_to_reraise():
+            # These are fatal errors or control exceptions that should stop the
+            # engine, so we re-raise them.
             raise
         except failure_exceptions_to_catch() as e:
+            # If the error was raised by Hypothesis-internal code, re-raise it
+            # as a fatal error instead of treating it as a test failure.
             escalate_hypothesis_internal_error()
+
             if data.frozen:
                 # This can happen if an error occurred in a finally
                 # block somewhere, suppressing our original StopTest.
                 # We raise a new one here to resume normal operation.
                 raise StopTest(data.testcounter)
             else:
+                # The test failed by raising an exception, so we inform the
+                # engine that this test run was interesting. This is the normal
+                # path for test runs that fail.
+
                 tb = get_trimmed_traceback()
                 info = data.extra_information
                 info.__expected_traceback = "".join(
@@ -635,7 +633,10 @@ class StateForActualGivenExecution(object):
                 lineno = origin[1]
                 data.mark_interesting((type(e), filename, lineno))
 
-    def run(self):
+    def run_engine(self):
+        """Run the test function many times, on database input and generated
+        input, using the Conjecture engine.
+        """
         # Tell pytest to omit the body of this function from tracebacks
         __tracebackhide__ = True
         if global_force_seed is None:
@@ -643,19 +644,16 @@ class StateForActualGivenExecution(object):
         else:
             database_key = None
         runner = ConjectureRunner(
-            self.evaluate_test_data,
+            self._execute_once_for_engine,
             settings=self.settings,
             random=self.random,
             database_key=database_key,
         )
-        try:
-            runner.run()
-        finally:
-            self.used_examples_from_database = runner.used_examples_from_database
+        # Use the Conjecture engine to run the test function many times
+        # on different inputs.
+        runner.run()
         note_engine_for_statistics(runner)
 
-        self.used_examples_from_database = runner.used_examples_from_database
-
         if runner.call_count == 0:
             return
         if runner.interesting_examples:
@@ -674,8 +672,12 @@ class StateForActualGivenExecution(object):
         if not self.falsifying_examples:
             return
         elif not self.settings.report_multiple_bugs:
+            # Pretend that we only found one failure, by discarding the others.
             del self.falsifying_examples[:-1]
 
+        # The engine found one or more failures, so we need to reproduce and
+        # report them.
+
         self.failed_normally = True
 
         flaky = 0
@@ -687,7 +689,7 @@ class StateForActualGivenExecution(object):
             self.__was_flaky = False
             assert info.__expected_exception is not None
             try:
-                self.execute(
+                self.execute_once(
                     ran_example,
                     print_example=True,
                     is_final=True,
@@ -704,9 +706,15 @@ class StateForActualGivenExecution(object):
                 )
             except BaseException as e:
                 if len(self.falsifying_examples) <= 1:
+                    # There is only one failure, so we can report it by raising
+                    # it directly.
                     raise
+
+                # We are reporting multiple failures, so we need to manually
+                # print each exception's stack trace and information.
                 tb = get_trimmed_traceback()
                 report("".join(traceback.format_exception(type(e), e, tb)))
+
             finally:  # pragma: no cover
                 # This section is in fact entirely covered by the tests in
                 # test_reproduce_failure, but it seems to trigger a lovely set
@@ -817,13 +825,21 @@ def given(
             test.__name__, original_argspec, generator_arguments, generator_kwargs
         )
 
+        # If the argument check found problems, return a dummy test function
+        # that will raise an error if it is actually called.
         if check_invalid is not None:
             return check_invalid
 
-        for name, strategy in zip(
-            reversed(original_argspec.args), reversed(generator_arguments)
-        ):
-            generator_kwargs[name] = strategy
+        # Because the argument check succeeded, we can convert @given's
+        # positional arguments into keyword arguments for simplicity.
+        if generator_arguments:
+            assert not generator_kwargs
+            for name, strategy in zip(
+                reversed(original_argspec.args), reversed(generator_arguments)
+            ):
+                generator_kwargs[name] = strategy
+        # These have been converted, so delete them to prevent accidental use.
+        del generator_arguments
 
         argspec = new_given_argspec(original_argspec, generator_kwargs)
 
@@ -852,6 +868,8 @@ def given(
 
             random = get_random_for_wrapped_test(test, wrapped_test)
 
+            # Use type information to convert "infer" arguments into appropriate
+            # strategies.
             if infer in generator_kwargs.values():
                 hints = get_type_hints(test)
             for name in [
@@ -868,7 +886,6 @@ def given(
                 wrapped_test,
                 arguments,
                 kwargs,
-                generator_arguments,
                 generator_kwargs,
                 argspec,
                 test,
@@ -905,6 +922,9 @@ def given(
 
             reproduce_failure = wrapped_test._hypothesis_internal_use_reproduce_failure
 
+            # If there was a @reproduce_failure decorator, use it to reproduce
+            # the error (or complain that we couldn't). Either way, this will
+            # always raise some kind of error.
             if reproduce_failure is not None:
                 expected_version, failure = reproduce_failure
                 if expected_version != __version__:
@@ -918,7 +938,7 @@ def given(
                         % (expected_version, __version__)
                     )
                 try:
-                    state.execute(
+                    state.execute_once(
                         ConjectureData.for_buffer(decode_failure(failure)),
                         print_example=True,
                         is_final=True,
@@ -940,10 +960,17 @@ def given(
                         "generated?"
                     )
 
+            # There was no @reproduce_failure, so start by running any explicit
+            # examples from @example decorators.
+
             execute_explicit_examples(
                 test_runner, test, wrapped_test, settings, arguments, kwargs
             )
 
+            # If there were any explicit examples, they all ran successfully.
+            # The next step is to use the Conjecture engine to run the test on
+            # many different inputs.
+
             if settings.max_examples <= 0:
                 return
 
@@ -956,13 +983,17 @@ def given(
                 if isinstance(runner, TestCase) and hasattr(runner, "subTest"):
                     subTest = runner.subTest
                     try:
-                        setattr(runner, "subTest", fake_subTest)
-                        state.run()
+                        runner.subTest = fake_subTest
+                        state.run_engine()
                     finally:
-                        setattr(runner, "subTest", subTest)
+                        runner.subTest = subTest
                 else:
-                    state.run()
+                    state.run_engine()
             except BaseException as e:
+                # The exception caught here should either be an actual test
+                # failure (or MultipleFailures), or some kind of fatal error
+                # that caused the engine to stop.
+
                 generated_seed = wrapped_test._hypothesis_internal_use_generated_seed
                 with local_settings(settings):
                     if not (state.failed_normally or generated_seed is None):
@@ -1000,6 +1031,9 @@ def given(
                     )
                     raise the_error_hypothesis_found
 
+        # After having created the decorated test function, we need to copy
+        # over some attributes to make the switch as seamless as possible.
+
         for attrib in dir(test):
             if not (attrib.startswith("_") or hasattr(wrapped_test, attrib)):
                 setattr(wrapped_test, attrib, getattr(test, attrib))
